@@ -46,15 +46,27 @@ typedef struct _OutputQueue
   GQueue        *queue;
 } OutputQueue;
 
+typedef void (*PushedCb) (OutputQueue *q, gpointer user_data, GError *error);
+
 typedef struct _OutputQueueElem
 {
   OutputQueue  *queue;
   const guint8 *buf;
   gsize         size;
-  GFunc         cb;
+  PushedCb      cb;
   gpointer      user_data;
 } OutputQueueElem;
 
+typedef struct _ServiceData
+{
+#ifdef G_OS_WIN32
+  gchar drive_letter;
+  GMutex mutex;
+#endif
+} ServiceData;
+
+static GCancellable *cancel;
+
 static OutputQueue*
 output_queue_new (GOutputStream *output)
 {
@@ -90,6 +102,8 @@ output_queue_ref (OutputQueue *q)
 static void
 output_queue_unref (OutputQueue *q)
 {
+  g_return_if_fail (q != NULL);
+
   q->refs--;
   if (q->refs == 0)
     output_queue_free (q);
@@ -143,26 +157,21 @@ output_queue_idle (gpointer user_data)
     }
 
   g_debug ("flushing %" G_GSIZE_FORMAT, e->size);
-  g_output_stream_write_all (q->output, e->buf, e->size, NULL, NULL, &error);
-  if (error)
-    goto end;
-
+  g_output_stream_write_all (q->output, e->buf, e->size, NULL, cancel, &error);
   if (e->cb)
-    e->cb (q, e->user_data);
+    e->cb (q, e->user_data, error);
+
+  if (error)
+      goto end;
 
   q->flushing = TRUE;
-  g_output_stream_flush_async (q->output, G_PRIORITY_DEFAULT, NULL, output_queue_flush_cb, e);
+  g_output_stream_flush_async (q->output, G_PRIORITY_DEFAULT, cancel, output_queue_flush_cb, e);
 
   q->idle_id = 0;
   return FALSE;
 
 end:
-  if (error)
-    {
-      g_warning ("error: %s", error->message);
-      g_clear_error (&error);
-    }
-
+  g_clear_error (&error);
   q->idle_id = 0;
   g_free (e);
   output_queue_unref (q);
@@ -172,10 +181,13 @@ end:
 
 static void
 output_queue_push (OutputQueue *q, const guint8 *buf, gsize size,
-                   GFunc pushed_cb, gpointer user_data)
+                   PushedCb pushed_cb, gpointer user_data)
 {
-  OutputQueueElem *e = g_new (OutputQueueElem, 1);
+  OutputQueueElem *e;
 
+  g_return_if_fail (q != NULL);
+
+  e = g_new (OutputQueueElem, 1);
   e->buf = buf;
   e->size = size;
   e->cb = pushed_cb;
@@ -204,18 +216,32 @@ typedef struct _Client
   OutputQueue       *queue;
 } Client;
 
+static gboolean quit_service;
 static GMainLoop *loop;
 static GInputStream *mux_istream;
 static GOutputStream *mux_ostream;
 static OutputQueue *mux_queue;
 static GHashTable *clients;
+static GSocketService *socket_service;
+#ifdef G_OS_UNIX
+static gint port_fd;
+#elif defined(G_OS_WIN32)
+static HANDLE port_handle;
+#endif
+
 static void start_mux_read (GInputStream *istream);
+#ifdef WITH_AVAHI
+static void mdns_unregister_service (void);
+#endif
 
 static void
 quit (int sig)
 {
   g_debug ("quit %d", sig);
 
+  if (sig == SIGINT || sig == SIGTERM)
+      quit_service = TRUE;
+
   g_main_loop_quit (loop);
 }
 
@@ -246,6 +272,9 @@ add_client (GSocketConnection *client_connection)
 static void
 client_free (Client *c)
 {
+  g_debug ("Free client %p", c);
+
+  g_io_stream_close (G_IO_STREAM (c->client_connection), NULL, NULL);
   g_object_unref (c->client_connection);
   output_queue_unref (c->queue);
   g_free (c);
@@ -256,7 +285,7 @@ remove_client (Client *client)
 {
   g_debug ("remove client %p", client);
 
-  g_hash_table_remove (clients, client->client_connection);
+  g_hash_table_remove (clients, &client->id);
 }
 
 typedef struct ReadData
@@ -283,9 +312,9 @@ read_thread (GSimpleAsyncResult *simple,
                            data->buffer, data->count, &bread,
                            cancellable, &error);
   g_debug ("my read result %" G_GSIZE_FORMAT, bread);
-  if (bread != data->count)
+  if (bread != data->count) {
     data->size = -1;
-  else
+  } else
     data->size = bread;
 
   if (error)
@@ -343,8 +372,26 @@ my_input_stream_read_finish (GInputStream *stream,
 }
 
 static void
-mux_pushed_client_cb (OutputQueue *q, gpointer user_data)
+handle_push_error (OutputQueue *q, gpointer user_data, GError *error)
 {
+  Client *client = user_data;
+
+  if (!error)
+    return;
+
+  g_warning ("push error: %s", error->message);
+  remove_client (client);
+  return;
+}
+
+static void
+mux_pushed_client_cb (OutputQueue *q, gpointer user_data, GError *error)
+{
+  if (error) {
+    handle_push_error (q, user_data, error);
+    return;
+  }
+
   start_mux_read (mux_istream);
 }
 
@@ -362,17 +409,17 @@ mux_data_read_cb (GObject      *source_object,
     {
       g_warning ("error: %s", error->message);
       g_clear_error (&error);
-      quit (1);
+      quit (-1);
       return;
     }
 
-  g_debug ("looking up client %" G_GINT64_FORMAT, demux.client);
   Client *c = g_hash_table_lookup (clients, &demux.client);
+  g_debug ("looked up client: %p", c);
   g_warn_if_fail(c != NULL);
 
   if (c)
     output_queue_push (c->queue, (guint8 *) demux.buf, demux.size,
-                       (GFunc) mux_pushed_client_cb, c);
+                       mux_pushed_client_cb, c);
   else
     start_mux_read (mux_istream);
 }
@@ -392,7 +439,7 @@ mux_size_read_cb (GObject      *source_object,
 
   my_input_stream_read_async (istream,
                               &demux.buf, demux.size, G_PRIORITY_DEFAULT,
-                              NULL, mux_data_read_cb, NULL);
+                              cancel, mux_data_read_cb, NULL);
   return;
 
 end:
@@ -402,7 +449,7 @@ end:
       g_clear_error (&error);
     }
 
-  quit (2);
+  quit (-2);
 }
 
 static void
@@ -420,7 +467,7 @@ mux_client_read_cb (GObject      *source_object,
     goto end;
   my_input_stream_read_async (istream,
                               &demux.size, sizeof (guint16), G_PRIORITY_DEFAULT,
-                              NULL, mux_size_read_cb, NULL);
+                              cancel, mux_size_read_cb, NULL);
   return;
 
 end:
@@ -430,7 +477,10 @@ end:
       g_clear_error (&error);
     }
 
-  quit (3);
+#ifdef WITH_AVAHI
+  mdns_unregister_service ();
+#endif
+  quit (-3);
 }
 
 static void
@@ -438,16 +488,21 @@ start_mux_read (GInputStream *istream)
 {
   my_input_stream_read_async (istream,
                               &demux.client, sizeof (gint64), G_PRIORITY_DEFAULT,
-                              NULL, mux_client_read_cb, NULL);
+                              cancel, mux_client_read_cb, NULL);
 }
 
 static void client_start_read (Client *client);
 
 static void
-mux_pushed_cb (OutputQueue *q, gpointer user_data)
+mux_pushed_cb (OutputQueue *q, gpointer user_data, GError *error)
 {
   Client *client = user_data;
 
+  if (error) {
+      handle_push_error (q, client, error);
+      return;
+  }
+
   if (client->size == 0)
     {
       remove_client (client);
@@ -467,7 +522,7 @@ client_read_cb (GObject      *source_object,
   gssize size;
 
   size = g_input_stream_read_finish (G_INPUT_STREAM (source_object), res, &error);
-  g_debug ("end read %" G_GSIZE_FORMAT, size);
+  g_debug ("end read %" G_GSSIZE_FORMAT, size);
   if (error)
     {
       g_warning ("error: %s", error->message);
@@ -480,9 +535,9 @@ client_read_cb (GObject      *source_object,
   g_return_if_fail (size >= 0);
   client->size = size;
 
-  output_queue_push (mux_queue, (guint8 *) &client->id, sizeof (gint64), NULL, NULL);
-  output_queue_push (mux_queue, (guint8 *) &client->size, sizeof (guint16), NULL, NULL);
-  output_queue_push (mux_queue, (guint8 *) client->buf, size, (GFunc) mux_pushed_cb, client);
+  output_queue_push (mux_queue, (guint8 *) &client->id, sizeof (gint64), handle_push_error, client);
+  output_queue_push (mux_queue, (guint8 *) &client->size, sizeof (guint16), handle_push_error, client);
+  output_queue_push (mux_queue, (guint8 *) client->buf, size, mux_pushed_cb, client);
 
   return;
 }
@@ -493,7 +548,7 @@ client_start_read (Client *client)
   GIOStream *iostream = G_IO_STREAM (client->client_connection);
   GInputStream *istream = g_io_stream_get_input_stream (iostream);
 
-  g_debug ("start read");
+  g_debug ("start read client %p", client);
   g_input_stream_read_async (istream,
                              client->buf, G_MAXUINT16, G_PRIORITY_DEFAULT,
                              NULL, client_read_cb, client);
@@ -516,6 +571,10 @@ incoming_callback (GSocketService    *service,
 
 static int port;
 
+#ifdef G_OS_WIN32
+static gboolean no_service;
+#endif
+
 #ifdef WITH_AVAHI
 static GaClient *mdns_client;
 static GaEntryGroup *mdns_group;
@@ -572,6 +631,24 @@ end:
   g_free (name);
 }
 
+static void
+mdns_unregister_service (void)
+{
+  GError *error = NULL;
+
+  if (mdns_group)
+    {
+      if (!ga_entry_group_reset (mdns_group, &error))
+        {
+          g_warning ("Could not disconnect MDNS service: %s", error->message);
+          g_clear_error (&error);
+        }
+
+      mdns_service = 0;
+      g_debug ("MDNS client disconected");
+    }
+}
+
 static void
 mdns_state_changed (GaClient *client, GaClientState state, gpointer user_data)
 {
@@ -589,11 +666,7 @@ mdns_state_changed (GaClient *client, GaClientState state, gpointer user_data)
     case GA_CLIENT_STATE_S_COLLISION:
     case GA_CLIENT_STATE_S_REGISTERING:
       g_message ("MDNS collision");
-      if (mdns_group)
-        {
-          ga_entry_group_reset (mdns_group, NULL);
-          mdns_service = 0;
-        }
+      mdns_unregister_service ();
       break;
 
     default:
@@ -643,44 +716,397 @@ open_mux_path (const char *path)
   g_return_if_fail (!mux_ostream);
   g_return_if_fail (!mux_queue);
 
+  g_debug ("Open %s", path);
 #ifdef G_OS_UNIX
-  gint fd = g_open (path, O_RDWR);
-  if (fd == -1)
+  port_fd = g_open (path, O_RDWR);
+  if (port_fd == -1)
       exit (1);
 
-  wait_for_virtio_host (fd);
+  wait_for_virtio_host (port_fd);
 
-  mux_ostream = g_unix_output_stream_new (fd, TRUE);
-  mux_istream = g_unix_input_stream_new (fd, FALSE);
+  mux_ostream = g_unix_output_stream_new (port_fd, TRUE);
+  mux_istream = g_unix_input_stream_new (port_fd, FALSE);
 #else
-  HANDLE h = CreateFile (path,
+  port_handle = CreateFile (path,
                          GENERIC_WRITE | GENERIC_READ,
                          0,
                          NULL,
                          OPEN_EXISTING,
                          FILE_FLAG_OVERLAPPED,
                          NULL);
-  g_assert (h != INVALID_HANDLE_VALUE);
 
-  mux_ostream = G_OUTPUT_STREAM (g_win32_output_stream_new (h, TRUE));
-  mux_istream = G_INPUT_STREAM (g_win32_input_stream_new (h, TRUE));
+  if (port_handle == INVALID_HANDLE_VALUE)
+      g_error ("%s", g_win32_error_message (GetLastError ()));
+
+  mux_ostream = G_OUTPUT_STREAM (g_win32_output_stream_new (port_handle, TRUE));
+  mux_istream = G_INPUT_STREAM (g_win32_input_stream_new (port_handle, TRUE));
 #endif
 
   mux_queue = output_queue_new (G_OUTPUT_STREAM (mux_ostream));
+}
+
+#ifdef G_OS_WIN32
+#define MAX_SHARED_FOLDER_NAME_SIZE 64
+#define MAX_DRIVE_LETTER_SIZE 3
+typedef enum _MapDriveEnum
+{
+  MAP_DRIVE_OK,
+  MAP_DRIVE_TRY_AGAIN,
+  MAP_DRIVE_ERROR
+} MapDriveEnum;
+
+typedef struct _MapDriveData
+{
+  ServiceData *service_data;
+  GCancellable *cancel_map;
+} MapDriveData;
+
+static gchar
+get_free_drive_letter(void)
+{
+  const guint32 max_mask = 1 << 25;
+  guint32 drives;
+  gint i;
+
+  drives = GetLogicalDrives ();
+  if (drives == 0)
+    {
+      g_warning ("%s", g_win32_error_message (GetLastError ()));
+      return 0;
+    }
+
+  for (i = 0; i < 26; i++)
+    {
+      guint32 mask = max_mask >> i;
+      if ((drives & mask) == 0)
+        return 'z' - i;
+    }
+
+  return 0;
+}
+
+static gchar
+get_spice_folder_letter(void)
+{
+  const guint32 max_mask = 1 << 25;
+  gchar local_name[3] = "z:";
+  gchar *spice_folder_name;
+  gchar spice_folder_letter = 0;
+  guint32 drives;
+  guint32 i;
+
+  drives = GetLogicalDrives ();
+  spice_folder_name = g_strdup_printf("\\\\localhost@%d\\DavWWWRoot", port);
+
+  for (i = 0; i < 26; i++)
+    {
+      gchar remote_name[MAX_SHARED_FOLDER_NAME_SIZE];
+      guint32 size = sizeof(remote_name);
+      guint32 mask = max_mask >> i;
+      if (drives & mask)
+        {
+          local_name[0] = 'z' - i;
+          if ((WNetGetConnection (local_name, remote_name, (LPDWORD)&size) == NO_ERROR) &&
+              (g_strcmp0 (remote_name, spice_folder_name) == 0))
+            {
+              spice_folder_letter = local_name[0];
+              g_debug ("Found Spice Shared Folder at %c drive", spice_folder_letter);
+              break;
+            }
+        }
+    }
+
+  g_free(spice_folder_name);
+
+  return spice_folder_letter;
+}
+
+/* User is required to call netresource_free, when no longer needed. */
+static void
+netresource_init(NETRESOURCE *net_resource, const gchar drive_letter)
+{
+  net_resource->dwType = RESOURCETYPE_DISK;
+  net_resource->lpLocalName = g_strdup_printf("%c:", drive_letter);
+  net_resource->lpRemoteName = g_strdup_printf("http://localhost:%d/", port);
+  net_resource->lpProvider = NULL;
+}
+
+static void
+netresource_free(NETRESOURCE *net_resource)
+{
+  g_free(net_resource->lpLocalName);
+  g_free(net_resource->lpRemoteName);
+}
+
+static MapDriveEnum
+map_drive(const gchar drive_letter)
+{
+  NETRESOURCE net_resource;
+  guint32 errn;
+
+  netresource_init(&net_resource, drive_letter);
+  errn = WNetAddConnection2 (&net_resource, NULL, NULL, CONNECT_TEMPORARY);
+  netresource_free(&net_resource);
+
+  if (errn == NO_ERROR)
+    {
+      g_debug ("Shared folder mapped to %c succesfully", drive_letter);
+      return MAP_DRIVE_OK;
+    }
+  else if (errn == ERROR_ALREADY_ASSIGNED)
+    {
+      g_debug ("Drive letter %c is already assigned", drive_letter);
+      return MAP_DRIVE_TRY_AGAIN;
+    }
+
+  g_warning ("map_drive error %d", errn);
+  return MAP_DRIVE_ERROR;
+}
+
+static void
+unmap_drive(ServiceData *service_data)
+{
+  gchar local_name[MAX_DRIVE_LETTER_SIZE];
+  guint32 errn;
+
+  g_mutex_lock(&service_data->mutex);
+  g_snprintf(local_name, MAX_DRIVE_LETTER_SIZE, "%c:", service_data->drive_letter);
+  errn = WNetCancelConnection2(local_name, CONNECT_UPDATE_PROFILE, TRUE);
+
+  if (errn == NO_ERROR)
+    {
+      g_debug ("Shared folder unmapped succesfully");
+    }
+  else if (errn == ERROR_NOT_CONNECTED)
+    {
+      g_debug ("Drive %c is not connected", service_data->drive_letter);
+    }
+  else
+    {
+      g_warning ("map_drive error %d", errn);
+    }
+
+  g_mutex_unlock(&service_data->mutex);
+  return;
+}
+
+static void
+map_drive_cb(GTask *task,
+             gpointer source_object,
+             gpointer task_data,
+             GCancellable *cancellable)
+{
+  const guint32 delay = 500; //half a second
+  MapDriveData *map_drive_data = task_data;
+  gchar drive_letter;
+  GPollFD cancel_pollfd;
+  guint32 ret = 0;
+
+  if (!g_cancellable_make_pollfd (map_drive_data->cancel_map, &cancel_pollfd))
+    {
+      g_critical ("GPollFD failed to create.");
+      return;
+    }
+
+  ret = g_poll (&cancel_pollfd, 1, delay);
+  g_cancellable_release_fd (map_drive_data->cancel_map);
+
+  if (ret != 0)
+    {
+      return;
+    }
+
+  while (TRUE)
+    {
+      drive_letter = get_free_drive_letter ();
+      if (drive_letter == 0)
+        {
+          g_warning ("all drive letters already assigned.");
+          break;
+        }
+
+      if (map_drive (drive_letter) != MAP_DRIVE_TRY_AGAIN)
+        {
+          break;
+        }
+      //TODO: After mapping, rename network drive from \\localhost@PORT\DavWWWRoot
+      //      to something like SPICE Shared Folder
+    }
+
+  g_mutex_lock(&map_drive_data->service_data->mutex);
+  map_drive_data->service_data->drive_letter = drive_letter;
+  g_mutex_unlock(&map_drive_data->service_data->mutex);
+}
+
+#endif
+
+static void
+run_service (ServiceData *service_data)
+{
+  g_debug ("Run service");
+
+#ifdef G_OS_WIN32
+  MapDriveData map_drive_data;
+  map_drive_data.cancel_map = g_cancellable_new ();
+  gchar drive_letter = get_spice_folder_letter ();
+
+  g_mutex_lock(&service_data->mutex);
+  service_data->drive_letter = drive_letter;
+  map_drive_data.service_data = service_data;
+  g_mutex_unlock(&service_data->mutex);
+
+  if (drive_letter == 0)
+    {
+      GTask *map_drive_task = g_task_new (NULL, NULL, NULL, NULL);
+      g_task_set_task_data (map_drive_task, &map_drive_data, NULL);
+      g_task_run_in_thread (map_drive_task, map_drive_cb);
+      g_object_unref (map_drive_task);
+    }
+#endif
+
+  g_socket_service_start (socket_service);
+
+  cancel = g_cancellable_new ();
+  clients = g_hash_table_new_full (g_int64_hash, g_int64_equal,
+                                   NULL, (GDestroyNotify) client_free);
+
+  loop = g_main_loop_new (NULL, TRUE);
+#ifdef G_OS_UNIX
+  open_mux_path ("/dev/virtio-ports/org.spice-space.webdav.0");
+#else
+  open_mux_path ("\\\\.\\Global\\org.spice-space.webdav.0");
+#endif
+
+  /* listen on port for incoming clients, multiplex there input into
+     virtio path, demultiplex input from there to the respective
+     clients */
+
+#ifdef WITH_AVAHI
+  GError *error = NULL;
+
+  mdns_client = ga_client_new (GA_CLIENT_FLAG_NO_FLAGS);
+  g_signal_connect (mdns_client, "state-changed", G_CALLBACK (mdns_state_changed), NULL);
+  if (!ga_client_start (mdns_client, &error))
+    {
+      g_printerr ("%s\n", error->message);
+      exit (1);
+    }
+#endif
 
   start_mux_read (mux_istream);
+  g_main_loop_run (loop);
+  g_main_loop_unref (loop);
+
+#ifdef G_OS_WIN32
+  g_cancellable_cancel (map_drive_data.cancel_map);
+  g_object_unref (map_drive_data.cancel_map);
+#endif
+
+  g_cancellable_cancel (cancel);
+
+  g_clear_object (&mux_istream);
+  g_clear_object (&mux_ostream);
+
+  output_queue_unref (mux_queue);
+  g_hash_table_unref (clients);
+
+  g_socket_service_stop (socket_service);
+
+  mux_queue = NULL;
+  g_clear_object (&cancel);
+
+#ifdef G_OS_WIN32
+  CloseHandle (port_handle);
+#else
+  close (port_fd);
+#endif
 }
 
+#ifdef G_OS_WIN32
+static SERVICE_STATUS service_status;
+static SERVICE_STATUS_HANDLE service_status_handle;
+
+DWORD WINAPI
+service_ctrl_handler(DWORD ctrl, DWORD type, LPVOID data, LPVOID ctx);
+VOID WINAPI
+service_main(DWORD argc, TCHAR *argv[]);
+
+DWORD WINAPI
+service_ctrl_handler (DWORD ctrl, DWORD type, LPVOID data, LPVOID ctx)
+{
+  DWORD ret = NO_ERROR;
+
+  ServiceData *service_data = ctx;
+
+  switch (ctrl)
+    {
+    case SERVICE_CONTROL_STOP:
+    case SERVICE_CONTROL_SHUTDOWN:
+        if (service_data->drive_letter != 0)
+          {
+            unmap_drive (service_data);
+            g_mutex_clear(&service_data->mutex);
+          }
+        quit (SIGTERM);
+        service_status.dwCurrentState = SERVICE_STOP_PENDING;
+        SetServiceStatus (service_status_handle, &service_status);
+        break;
+
+    default:
+        ret = ERROR_CALL_NOT_IMPLEMENTED;
+    }
+
+  return ret;
+}
+
+VOID WINAPI
+service_main (DWORD argc, TCHAR *argv[])
+{
+  ServiceData service_data;
+
+  service_data.drive_letter = 0;
+  g_mutex_init(&service_data.mutex);
+
+  service_status_handle =
+    RegisterServiceCtrlHandlerEx ("spice-webdavd", service_ctrl_handler, &service_data);
+
+  g_return_if_fail (service_status_handle != 0);
+
+  service_status.dwServiceType = SERVICE_WIN32;
+  service_status.dwCurrentState = SERVICE_RUNNING;
+  service_status.dwControlsAccepted = SERVICE_ACCEPT_STOP | SERVICE_ACCEPT_SHUTDOWN;
+  service_status.dwWin32ExitCode = NO_ERROR;
+  service_status.dwServiceSpecificExitCode = NO_ERROR;
+  service_status.dwCheckPoint = 0;
+  service_status.dwWaitHint = 0;
+  SetServiceStatus (service_status_handle, &service_status);
+
+  while (!quit_service) {
+      run_service (&service_data);
+      g_usleep (G_USEC_PER_SEC);
+  }
+
+  service_status.dwCurrentState = SERVICE_STOPPED;
+  SetServiceStatus (service_status_handle, &service_status);
+}
+#endif
+
 static GOptionEntry entries[] = {
   { "port", 'p', 0,
     G_OPTION_ARG_INT, &port,
     "Port to listen on", NULL },
+#ifdef G_OS_WIN32
+  { "no-service", 0, 0,
+    G_OPTION_ARG_NONE, &no_service,
+    "Don't start as a service", NULL },
+#endif
   { NULL }
 };
 
 int
 main (int argc, char *argv[])
 {
+  ServiceData service_data;
   GOptionContext *opts;
   GError *error = NULL;
 
@@ -703,13 +1129,16 @@ main (int argc, char *argv[])
 
   signal (SIGINT, quit);
 
-
-  GSocketService *service = g_socket_service_new ();
-  GInetAddress *iaddr = g_inet_address_new_loopback (G_SOCKET_FAMILY_IPV4);
+  /* run socket service once at beginning, there seems to be a bug on
+     windows, and it can't accept new connections if cleanup and
+     restart a new service */
+  socket_service = g_socket_service_new ();
+  GInetAddress *iaddr = g_inet_address_new_any (G_SOCKET_FAMILY_IPV4);
   GSocketAddress *saddr = g_inet_socket_address_new (iaddr, port);
   g_object_unref (iaddr);
 
-  g_socket_listener_add_address (G_SOCKET_LISTENER (service), saddr,
+  g_socket_listener_add_address (G_SOCKET_LISTENER (socket_service),
+                                 saddr,
                                  G_SOCKET_TYPE_STREAM,
                                  G_SOCKET_PROTOCOL_TCP,
                                  NULL,
@@ -721,42 +1150,41 @@ main (int argc, char *argv[])
       exit (1);
     }
 
-  g_signal_connect (service,
+  g_signal_connect (socket_service,
                     "incoming", G_CALLBACK (incoming_callback),
                     NULL);
 
-  clients = g_hash_table_new_full (g_int64_hash, g_int64_equal,
-                                   NULL, (GDestroyNotify) client_free);
+#ifdef G_OS_WIN32
+  service_data.drive_letter = 0;
+  g_mutex_init(&service_data.mutex);
 
-  loop = g_main_loop_new (NULL, TRUE);
-#ifdef G_OS_UNIX
-  open_mux_path ("/dev/virtio-ports/org.spice-space.webdav.0");
-#else
-  open_mux_path ("\\\\.\\Global\\org.spice-space.webdav.0");
-#endif
-
-
-  /* listen on port for incoming clients, multiplex there input into
-     virtio path, demultiplex input from there to the respective
-     clients */
-
-  g_socket_service_start (service);
-
-#ifdef WITH_AVAHI
-  mdns_client = ga_client_new (GA_CLIENT_FLAG_NO_FLAGS);
-  g_signal_connect (mdns_client, "state-changed", G_CALLBACK (mdns_state_changed), NULL);
-  if (!ga_client_start (mdns_client, &error))
+  SERVICE_TABLE_ENTRY service_table[] =
     {
-      g_printerr ("%s\n", error->message);
-      exit (1);
+      { (char *)"spice-webdavd", service_main }, { NULL, NULL }
+    };
+  if (!no_service && !getenv("DEBUG"))
+    {
+      if (!StartServiceCtrlDispatcher (service_table))
+        {
+          g_error ("%s", g_win32_error_message (GetLastError ()));
+          exit (1);
+        }
+    } else
+#endif
+  while (!quit_service) {
+    run_service (&service_data);
+    g_usleep (G_USEC_PER_SEC);
+  }
+
+#ifdef G_OS_WIN32
+  if (service_data.drive_letter != 0)
+    {
+      unmap_drive (&service_data);
+      g_mutex_clear(&service_data.mutex);
     }
 #endif
 
-  g_main_loop_run (loop);
-  g_main_loop_unref (loop);
-
-  output_queue_unref (mux_queue);
-  g_hash_table_unref (clients);
+  g_clear_object (&socket_service);
 
   return 0;
 }
